""" ExtFile product module """
###############################################################################
#
# Copyright (c) 2001 Gregor Heine <mac.gregor@gmx.de>. All rights reserved.
# ExtFile Home: http://www.zope.org/Members/MacGregor/ExtFile/index_html
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission
#
# Disclaimer
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#   
#  In accordance with the license provided for by the software upon
#  which some of the source code has been derived or used, the following
#  acknowledgement is hereby provided :
#
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
###############################################################################

__doc__ = """
              ExtFile product module.
              The ExtFile-Product works like the Zope File-product, but stores 
              the uploaded file externally in a repository-direcory.
"""
__version__='$Release: 2.0.0 $'[10:-2]

import Globals
from __main__ import *
import Zope
from Products.ZCatalog.CatalogAwareness import CatalogAware
from OFS.SimpleItem import SimpleItem
from OFS import Image
from OFS.PropertyManager import PropertyManager
from Globals import DTMLFile, HTMLFile, MessageDialog
from OFS.content_types import guess_content_type
from webdav.common import rfc1123_date
from zLOG import LOG, ERROR, INFO, BLATHER, DEBUG

from DateTime import DateTime
import urllib, os, types, string
from os.path import join, isfile, split, splitext
try: from cStringIO import StringIO
except: from StringIO import StringIO
from shutil import copy2
import sys, traceback

FLAT = 0
SYNC_ZODB = 1
SLICED = 2

REPOSITORY = SYNC_ZODB

# format for the files in the repository:
# %u=user, %p=path, %n=file name, %e=file extension, %c=counter, %t=time
FILE_FORMAT = "%n%c%e"

BACKUP_ON_DELETE = 0
ALWAYS_BACKUP = 1
UNDO_POLICY = ALWAYS_BACKUP

#if this is true, the file in the repository will follow the
# id name. This is not working perfectly: it works on copy/paste,
# but not on rename
REPOSITORY_FOLLOW_ID_FLAG = 1

bad_chars =  ' ,;()[]{}ݟ'
good_chars = '_________AAAAAAaaaaaaCcEEEEEeeeeeIIIIiiiiNnOOOOOOooooooSssUUUUuuuuYYyyZz'
TRANSMAP = string.maketrans(bad_chars, good_chars)

manage_addExtFileForm = DTMLFile('objectAdd', globals(),Kind='ExtFile',kind='extfile')


def manage_addExtFile(self,id,file='',title='',precondition='',
		      content_type='',REQUEST=None, permission_check=0):
	""" Add a ExtFile to a folder. """
	
	id=str(id)
	title=str(title)
	content_type=str(content_type)
	precondition=str(precondition)


	id, title = Image.cookId(id, title, file)
	id = string.translate(id, TRANSMAP)
	
	self = self.this()
	
	# First, we create the file without data:

	tempExtFile = ExtFile(id,title,None,content_type, precondition)
	self._setObject(id, tempExtFile)
	self._getOb(id).manage_file_upload(file, content_type)
	
	if REQUEST is not None:
		REQUEST['RESPONSE'].redirect(self.absolute_url()+'/manage_main')

class ExtFile(CatalogAware, SimpleItem, PropertyManager): 
	
	# what properties have we?
	_properties=(
		{'id':'title',	'type':'string'},
		{'id':'descr', 	'type':'string'},
		{'id':'precondition', 	'type':'string'},
		{'id':'content_type',		'type':'string'},
		{'id':'use_download_permission_check',	'type':'int'},
		{'id':'filename', 'type': 'lines', 'mode': ''},

	)
	
	# what management options are there?
	manage_options = (
		{'label':'Edit',	'action': 'manage_main'		},
		{'label':'View/Download', 'action': ''						},
		{'label':'Upload',	'action': 'manage_uploadForm'	},
		{'label':'Properties',	'action': 'manage_propertiesForm' },
		{'label':'Security',	'action': 'manage_access'	},
	) 
	
	# what permissions make sense for us? 
	__ac_permissions__=(
		('View management screens',	('manage_tabs',
						 'manage_main',
						 'manage_uploadForm')),
		('Change permissions',		('manage_access',)),
		('Change ExtFile/ExtImage',	('manage_editExtFile',
						 'manage_edit',
						 'manage_file_upload',
						 'manage_upload',
						 'manage_http_upload',
						 'PUT')),
		('FTP access',			('manage_FTPstat',
						 'manage_FTPget',
						 'manage_FTPlist')),
		('Download ExtFile/ExtImage',()),
		                                  ('View',
						  ('index_html',
						   'icon_html',
						   'icon_gif',
						   'link',
						   'is_broken',
						   'get_size',
						   'getContentType',
						   '__str__'
						   )),
		)
	
	# what do people think they're adding? 
	meta_type = 'ExtFile'

	precondition=''
	
	# location of the file-repository
	_repository = ['var','reposit']
	
	use_download_permission_check = 0
	
	# MIME-Type Dictionary. To add a MIME-Type, add a file in the directory 
	# icons/_category_/_subcategory-icon-file_
	# example: Icon tifficon.gif for the MIME-Type image/tiff goes to 
	# icons/image/tifficon.gif and the dictionary must be updated like this: 
	# 'image':{'tiff':'tifficon.gif','default':'default.gif'}, ...
	_types={'image':					
		{'default':'default.gif'},
		'text':
		{'html':'html.gif', 'xml':'xml.gif', 'default':'default.gif', 
		 'python':'py.gif'},
		'application':
		{'pdf':'pdf.gif', 'zip':'zip.gif', 'tar':'zip.gif', 
		 'msword':'doc.gif', 'excel':'xls.gif', 'powerpoint':'ppt.gif', 
		 'default':'default.gif'},
		'video':
		{'default':'default.gif'},
		'audio':
		{'default':'default.gif'},
		'default':'default.gif'
		}

	#"filename"s of backups of file (for undo)
	# actually, filenames are lists in this class; lists starting from instance_home
	_filebackups = []
	
	################################
	# Init method                  #
	################################
	
	def __init__(self, id, title, file, content_type='', precondition='', permission_check=0):
		#(self, id, title='', descr='', permission_check=0): 
		""" initialize a new instance of ExtFile """
		self.id = id
		self.title = title

		#this is not a property in OFS.Image! and it is a duplicate of DublinCore
		self.descr = ''

		self.use_download_permission_check = permission_check
		self.__version__ = __version__
		self.filename = []
		self.content_type = content_type
		self.precondition = precondition

		self._filebackups = []
		
		#this does not work, since absolute_url is unavailable
		if file:
			zLOG.LOG('ExtFile',ERROR, 'cant upload file here')
		# self.manage_file_upload(file, content_type, REQUEST=None)
			
	################################
	# Public methods               #
	################################

	#hack use .title
	def __str__(self):
		r= self.title
		try:
			r= self.index_html()
		except:
			LOG('ExtFile, error in __str__ :',ERROR,
			    '\n'.join(traceback.format_exception
				      (*sys.exc_info())))
		return r
	
	def __len__(self): return 1

	#HACK to make it more similar OFS.Image.File
	def index_html (self,REQUEST=None,RESPONSE=None,
			icon=0, preview=0, width=None, height=None, 
	                ):
		""" return the file with it's corresponding MIME-type """
		if REQUEST and RESPONSE == None:
			RESPONSE=REQUEST.RESPONSE
		# HTTP If-Modified-Since header handling. (Copied from OFS/Image.py)
		if REQUEST is None and hasattr(self,'REQUEST'): 
			REQUEST = self.REQUEST
		if REQUEST is not None and RESPONSE is not None :
			header = REQUEST.get_header('If-Modified-Since', None)
			if header is not None:
				header = string.split(header, ';')[0]
				try:    mod_since = long(DateTime(header).timeTime())
				except: mod_since = None
				if mod_since is not None:
					if self._p_mtime:
						last_mod = long(self._p_mtime)
					else:
						last_mod = long(0)
					if last_mod > 0 and last_mod < mod_since:
						RESPONSE.setStatus(304)
						return ''


		if self.precondition and hasattr(self,self.precondition):
		    # Grab whatever precondition was defined and then
		    # execute it.  The precondition will raise an exception
		    # if something violates its terms.
		    c=getattr(self,self.precondition)
		    if hasattr(c,'isDocTemp') and c.isDocTemp:
			c(REQUEST['PARENTS'][1],REQUEST)
		    else:
			c()

		#TODO: range support
		
		if hasattr(self,'has_preview') and self.has_preview: has_preview = 1
		else: has_preview = 0
		if not self._access_permitted(REQUEST): preview = 1
		if (preview and not has_preview): icon = 1
		
		if icon:
			filename = join(SOFTWARE_HOME, 'Products', 
			                'ExtFile', self.getIconPath())
			content_type = 'image/gif'
		elif preview:
			filename = self._get_filename(self.prev_filename)
			content_type = self.prev_content_type
		else:
			filename = self._get_filename(self.filename)
			content_type = self.content_type
		
		cant_read_exc = "Can't read: "
		if filename and not isfile(filename): self._undo()
		if filename and isfile(filename):
			size = os.stat(filename)[6] # file size
		else:
			filename = join(SOFTWARE_HOME, 'Products',
			              'ExtFile', 'icons', 'broken.gif')
			try: size = os.stat(filename)[6]
			except: raise cant_read_exc, ("%s (%s)" %(self.id, filename))
			content_type = 'image/gif'
			icon = 1
		data = StringIO()
		if icon==0 and width is not None and height is not None:
			try:
				from PIL import Image
				im = Image.open(filename) 
				if im.mode!='RGB' and im.mode!='CMYK': im = im.convert("RGB")
				im.draft(None,(int(width),int(height)))
				im.load()
				im = im.resize((int(width),int(height)), Image.BICUBIC)
				im.save(data, 'JPEG')
			except:
				self._copy(filename, data)
			else:
				data.seek(0,2)
				size = data.tell()
				content_type = 'image/jpeg'
		else:
			self._copy(filename, data)
		if RESPONSE is not None:
			last_mod = rfc1123_date(self._p_mtime)
			RESPONSE.setHeader('Last-Modified', last_mod)
			RESPONSE.setHeader('Content-Type', content_type)
			RESPONSE.setHeader('Content-Length', size)
		return data.getvalue()
	
	def view_image_or_file(self):
		""" The default view of the contents of the File or Image. """
		raise 'Redirect', self.absolute_url()

	def link(self, text='', **args):
		""" return a HTML link tag to the file """
		if text=='': text = self.title_or_id()
		strg = '<a href="%s"' % (self.absolute_url())
		for key in args.keys():
			value = args.get(key)
			strg = '%s %s="%s"' % (strg, key, value)
		strg = '%s>%s</a>' % (strg, text)
		return strg
	
	def icon_gif(self):
		""" return an icon for the file's MIME-Type """
		raise 'Redirect', self.absolute_url()+'?icon=1'
	
	def icon_html(self):
		""" The icon embedded in html with a link to the real file """
		return '<img src="%s?icon=1" border="0">' % self.absolute_url()
	
	def is_broken(self):
		""" Check if external file exists and return true (1) or false (0) """
		fn = self._get_filename(self.filename)
		if not filename :
			return 0
		if not isfile(fn):
			self._undo()
			if not isfile(fn):
				return 1
		return 0
	
	def get_size(self):
		""" Returns the size of the file or image """
		fn = self._get_filename(self.filename)
		if fn and not isfile(fn): self._undo()
		if fn and isfile(fn): size = os.stat(fn)[6]
		else: size = 0
		return size
	
	rawsize = get_size
	getSize = get_size
	
	def size(self):
		""" Returns a formatted stringified version of the file size """
		return self._bytetostring(self.get_size())
	
	def getContentType(self):
		""" Get the content type of a file or image.
		    Returns the content type (MIME type) of a file or image.
		"""
		return self.content_type
		
	def getIconPath(self):
		""" Depending on the MIME Type of the file/image an icon
		    can be displayed. This function determines which
		    image in the lib/python/Products/ExtFile/icons/...
		    directory shold be used as icon for this file/image
		"""
		cat, sub = self._getMIMECatAndSub(self.content_type)
		if hasattr(self,'has_preview') and self.has_preview: cat = 'image'
		if self._types.has_key(cat):
			file = self._types[cat]['default']
			for item in self._types[cat].keys():
				if string.find(sub,item)>=0:
					file = self._types[cat][item]
					break
			return join('icons',cat,file)
		else:
			return join('icons',self._types['default'])
	
	################################
	# Protected management methods #
	################################
	
	# Management Interface
	manage_main = Image.File.manage_main
	#HTMLFile('extFileEdit', globals())	
	
	def manage_editExtFile(self, title='', descr='', REQUEST=None): 
		""" Manage the edited values """
		if self.title!=title: self.title = title
		if self.descr!=descr: self.descr = descr
		# update ZCatalog
		self.reindex_object()
		if REQUEST is not None:
			return MessageDialog(
				title = 'Edited',
				message = "The properties of %s have been changed!" % self.id,
				action = './manage_main',
			)

	#as the OFS.Image
	def manage_edit(self, title, content_type, precondition=None,
                    filedata=None, REQUEST=None):
		"edit file, as in OFS.Image.File"
		if self.precondition is not None and self.precondition!=precondition:
			self.precondition=precondition
		if self.content_type!=content_type:
			self.content_type=content_type
		if filedata:
			self._backup(os.rename)
			try:
				outfile=self._get_filename()
				outstream = open(outfile, 'wb')
				outstream.write(filedata)
				outstream.close()
			except IOError:
				self._undo()
				
		return self.manage_editExtFile(title,self.descr,REQUEST)
	
	# File upload Interface
	manage_uploadForm = HTMLFile('extFileUpload', globals())
	
	def manage_file_upload(self, file='', content_type='', REQUEST=None):
		""" Upload file from local directory """
		new_fn = self._backup_and_get_filename()
		try:
			self._copy(file, self._get_filename(new_fn))
			self.content_type = self._get_content_type(file,
								   file.read(100), 
								   self.id,
								   content_type or self.content_type)
		except:
			if UNDO_POLICY==ALWAYS_BACKUP and self.filename: 
				self._undo()
			raise
		else:
			self.filename = new_fn
			
		if REQUEST is not None:
			return MessageDialog(title = 'Uploaded',
				message = "The file was uploaded successfully!", 
				action = './manage_main',)

	#as the OFS.Image
	def manage_upload(self,file,REQUEST=None):
		""" Upload file from local directory , as OFS.Image.File """
		content_type=''
		return self.manage_file_upload(file, content_type, REQUEST)
	

	
	def manage_http_upload(self, url, REQUEST=None):
		""" Upload file from http-server """
		new_fn = self._backup_and_get_filename()
		try:
			url = urllib.quote(url,'/:')
			fp_in = urllib.urlopen(url)
			self._copy(fp_in, self._get_filename(new_fn))
			instream = open(self._get_filename(new_fn), 'rb')
			self.content_type = self._get_content_type(instream, 
			                    instream.read(100),
			                    self.id, self.content_type)
			instream.close()
		except:
			if UNDO_POLICY==ALWAYS_BACKUP and self.filename: 
				self._undo()
			raise
		else:
			self.filename = new_fn
		if REQUEST is not None:
			return MessageDialog(title = 'Uploaded',
				message = "The file was uploaded successfully!", 
				action = './manage_main',)

	manage_FTPget = index_html
	
	def PUT(self, REQUEST, RESPONSE):
		""" Handle HTTP PUT requests """
		self.dav__init(REQUEST, RESPONSE)
		content_type = REQUEST.get_header('content-type', None)
		instream = REQUEST['BODYFILE']
		new_fn = self._backup_and_get_filename()
		try:
			self._copy(instream, self._get_filename(new_fn))
			self.content_type = self._get_content_type(instream, 
			                    instream.read(100),
			                    self.id, content_type or self.content_type)
		except:
			if UNDO_POLICY==ALWAYS_BACKUP and self.filename: 
				self._undo()
		else:
			self.filename = new_fn
		RESPONSE.setStatus(204)
		return RESPONSE
	
	################################
	# Private methods              #
	################################
	
	def _access_permitted(self, REQUEST):
		""" check if the user is allowed to download the file """
		if hasattr(self,'use_download_permission_check') and \
		   self.use_download_permission_check and \
		   (REQUEST is None or not REQUEST.has_key('AUTHENTICATED_USER') or 
		    not REQUEST['AUTHENTICATED_USER'].has_permission(
		                                'Download ExtFile/ExtImage', self)
		   ):
			return 0
		else: 
			return 1
	
	def _get_content_type(self, file, body, id, content_type=None):
		""" determine the mime-type """
		headers = getattr(file, 'headers', None)
		if headers and headers.has_key('content-type'):
			content_type = headers['content-type']
		else:
			if type(body) is not type(''): body = body.data
			content_type, enc = guess_content_type(getattr(file,'filename',id),
			                                       body, content_type)
		return content_type
	
	def _getMIMECatAndSub(self, mime_string):
		""" Split MIME String into Category and Subcategory """
		cat = mime_string[:string.find(mime_string, '/')]   # MIME-category
		sub = mime_string[string.find(mime_string, '/')+1:] # sub-category
		return cat, sub
	
	def _copy(self, infile, outfile):
		""" read binary data from infile and write it to outfile
		    infile and outfile my be strings, in which case a file with that
		    name is opened, or filehandles, in which case they are accessed
		    directly.
		"""
		if type(infile) is types.StringType: 
			try:
				instream = open(infile, 'rb')
			except IOError:
				#undo is managed outside self._undo()
				try:
					instream = open(infile, 'rb')
				except IOError:
					raise IOError, ("%s (%s)" %(self.id, infile))
			close_in = 1
		else:
			instream = infile
			close_in = 0
		if type(outfile) is types.StringType: 
			try:
				outstream = open(outfile, 'wb')
			except IOError:
				raise IOError, ("%s (%s)" %(self.id, outfile))
			close_out = 1
		else:
			outstream = outfile
			close_out = 0
		try:
			blocksize = 2<<16
			block = instream.read(blocksize)
			outstream.write(block)
			while len(block)==blocksize:
				block = instream.read(blocksize)
				outstream.write(block)
		except IOError:
			raise IOError, ("%s (%s)" %(self.id, filename))
		try: instream.seek(0)
		except: pass
		if close_in: instream.close()
		if close_out: outstream.close()
	
	def _bytetostring (self, value):
		""" Convert an int-value (file-size in bytes) to an String
		    with the file-size in Byte, KB or MB
		"""
		bytes = float(value)
		if bytes>=1000:
			bytes = bytes/1024
			if bytes>=1000:
				bytes = bytes/1024
				typ = ' MB'
			else:
				typ = ' KB'
		else:
			typ = ' Bytes'
		strg = '%4.2f'%bytes
		strg = strg[:4]
		if strg[3]=='.': strg = strg[:3]
		strg = strg+typ
		return strg
		
	def _undo (self):
		""" restore file after undo or copy-paste """
		if self._filebackups and self.filename :
			fn_path = self._get_filename()
			fn_undo = self._filebackups.pop()
			fn_undo_path = self._get_filename(fn_undo)
			if  isfile(fn_undo_path):
				os.rename(fn_undo_path, fn_path)
				LOG("ExtFile",  DEBUG," '%s': undo '%s' from '%s' " %
				    (self.id,fn_path,fn_undo_path) )
			else:
				LOG("ExtFile" , ERROR,
				    "can't undo '%s' from '%s' : no such file" %
				    (self.id, fn_undo_path))


	def _get_unique_undo_filename (self, file=None):
		if file == None:
			file = self.filename[:]
		dirs = file[:-1]
		path, base = split(self._get_filename(file))
		fileformat = base + '.%c.undo'
		dirs.append(self._search_unique(path,fileformat))
		return dirs

	def _backup (self, action):
		""" store file for undo or copy-paste """
		if self.filename:
			fn = self._get_filename(self.filename)
			fnup = self._get_unique_undo_filename()
			fnu = self._get_filename(fnup)
			try:	
				action(fn, fnu)
			except OSError, msg:
				LOG("ExtFile can't backup '%s':" %self.id,
				    ERROR, msg)
			else:
				LOG("ExtFile", DEBUG, " '%s': backup '%s' to '%s' " %
				    (self.id,fn,fnu) )
				self._filebackups.append(fnup)

	def __del__ (self):
		" deletes files from filesystem "
		while self._filebackups:
			f = self._filebackups.pop()
			fn = self._get_filename(f)
			if isfile(fn):
				LOG('ExtFile', BLATHER,
				    ' %s: deleting undo file %s' %
				    (self.id,fn))
				os.remove(fn)
		fn = self._get_filename()
		## this is dangeroues
		if fn and isfile(fn):
			LOG("ExtFile" , PROBLEM,
			    "the class '%s' was deleted but the file '%s' is still there" %   (self.id, fn))

		#	LOG('ExtFile %s' %self.id, BLATHER, 'deleting %s' % fn)
		#	os.remove(self.filename)
		#	self.filename=[]
			
	def _get_filename(self, filename=None):
		""" Generate the full filename, incuding directories from 
		    self._repository and self.filename
		"""
		if filename == None :
			filename = self.filename
		if filename == None or filename == [] :
			return None
		path = INSTANCE_HOME
		for item in self._repository:
			path = join(path,item)
		if type(filename)==types.ListType:
			for item in filename:
				path = join(path,item)
		elif type(filename)==types.StringType :
			path = join(path,filename)
		else:
			raise Exception, 'program error'
		return path

	##I dont understands what this function does:
	## it returns a filename but actually does a backup, but maybe undoes it !?!?
	
## 	def _get_ufn(self, filename):
## 		""" If no unique filename has been generated, generate one
## 		    otherwise, return the existing one.
## 		"""
## 		if UNDO_POLICY==ALWAYS_BACKUP or filename==[]: 
## 			new_fn = self._get_new_ufn()
## 		else: 
## 			new_fn = filename[:]
## 		if filename:
## 			old_fn = self._get_filename(filename)
## 			if UNDO_POLICY==ALWAYS_BACKUP: 
## 				try: os.rename(old_fn, old_fn+'.undo')
## 				except: pass
## 			else:
## 				try: os.rename(old_fn+'.undo', old_fn)
## 				except: pass
## 		return new_fn

	## so I have created a new function in its place...

	def _backup_and_get_filename(self):
		if UNDO_POLICY==ALWAYS_BACKUP and self.filename:
			self._backup (os.rename)  #(copy2)
		if self.filename and not REPOSITORY_FOLLOW_ID_FLAG :
			new_fn=self.filename
		else:
			new_fn=self._get_new_ufn()
		return new_fn


	def _get_new_ufn(self):
		""" Create a new unique filename """
		rel_url_list = string.split(self.absolute_url(1), '/')[:-1]
		rel_url_list = filter(None, rel_url_list)

		id_name, id_ext =  splitext(self.id)
		## this deletes the dot from the extension
		#if id_ext:
		#	id_ext = id_ext[1:]
		
		# generate directory structure
		dirs = []
		if REPOSITORY==SYNC_ZODB: 
			dirs = rel_url_list
		elif REPOSITORY==SLICED: 
			slice_depth = 2 # modify here, if you want a different slice depth
			slice_width = 1 # modify here, if you want a different slice width
			temp = id_name
			for i in range(slice_depth):
				if len(temp)<slice_width*(slice_depth-i): 
					dirs.append(slice_width*'_')
				else:
					dirs.append(temp[:slice_width])
					temp=temp[slice_width:]
		
		# generate file format
		# time/counter (%t)
		fileformat = FILE_FORMAT
		if string.find(fileformat, "%t")>=0:
			fileformat = string.replace(fileformat, "%t", "%c")
			counter = int(DateTime().strftime('%m%d%H%M%S'))
		else:
			counter = 0

		# user (%u)
		if string.find(fileformat, "%u")>=0:
			if (hasattr(self, "REQUEST") and 
			   self.REQUEST.has_key('AUTHENTICATED_USER')):
				user = self.REQUEST['AUTHENTICATED_USER'].name 
				fileformat = string.replace(fileformat, "%u", user)
			else:
				fileformat = string.replace(fileformat, "%u", "")
		# path (%p)
		if string.find(fileformat, "%p")>=0:
			temp = string.joinfields (rel_url_list, "_")
			fileformat = string.replace(fileformat, "%p", temp)
		# file and extension (%n and %e)
		if string.find(fileformat,"%n")>=0 or string.find(fileformat,"%e")>=0:
			fileformat = string.replace(fileformat, "%n", id_name)
			fileformat = string.replace(fileformat, "%e", id_ext)
		
		# make the directories
		path = self._get_filename(dirs)
		if not os.path.isdir(path):
			mkdir_exc = "Can't create directory: "
			try:
				os.makedirs(path)
			except:
				raise mkdir_exc, path
			
		file = self._search_unique(path,fileformat,counter)
		dirs.append(file)
		return dirs

	def _search_unique(self, path, fileformat, counter=0):
		""" uses a %c in the file for creating an unique filename
		return the string fileformat, substituted """

		invalid_format_exc = "Invalid file format: "
		if string.find(fileformat, "%c") < 0:
			raise invalid_format_exc, fileformat
			
		# search for unique filename
		while 1:
			if counter: 
				f = string.replace(fileformat, "%c", `counter`)
			else: 
				f = string.replace(fileformat, "%c", '')
			fn = join(path, f)
		
			if (isfile(fn) or isfile(fn+'.preview') 
			    or isfile(fn+'.undo') or isfile(fn+'.preview.undo')):
				counter = counter + 1
			else:
				break

		return f
	
	################################
	# Special management methods   #
	################################
	
	def manage_afterClone(self, item, new_fn=None):
		""" When a copy of the object is created (zope copy-paste-operation),
		    this function is called by CopySupport.py. A copy of the external 
		    file is created and self.filename is changed.
		"""
		try: 
			self.absolute_url(1) # this raises an exception, if no context
		except: 
			self._v_has_been_cloned=1
		else:
			old_fn = self._get_filename(self.filename)
			new_fn = new_fn or self._get_new_ufn()
			if isfile(old_fn):
				self._copy(old_fn, self._get_filename(new_fn))
			self.filename = new_fn
		return ExtFile.inheritedAttribute ("manage_afterClone") (self, item)
		
	def manage_afterAdd(self, item, container):
		""" This method is called, whenever _setObject in ObjectManager gets 
		    called. This is the case after a normal add and if the object is a 
		    result of cut-paste- or rename-operation. In the first case, the
		    external files doesn't exist yet, otherwise it was renamed to .undo
		    by manage_beforeDelete before and must be restored by _undo().
		"""
		self._undo()
		if hasattr(self, "_v_has_been_cloned"):
			delattr(self, "_v_has_been_cloned")
			self.manage_afterClone(item)
		return ExtFile.inheritedAttribute ("manage_afterAdd") \
		       (self, item, container)
	
	def manage_beforeDelete(self, item, container):
		""" This method is called, when the object is deleted. To support 
		    undo-functionality and because this happens too, when the object 
		    is moved (cut-paste) or renamed, the external file is not deleted. 
		    It are just renamed to filename.undo and remains in the 
		    repository, until it is deleted manually.
		"""
		if self.filename:
			self._backup (os.rename)
		return ExtFile.inheritedAttribute ("manage_beforeDelete") \
		       (self, item, container)

Globals.default__class_init__(ExtFile)
